// Copyright (C) 2021 Harry Godden (hgn)

// Low level disk reading
//=======================================================================================================================

// Read binary, or text assets. _s variants also give size in bytes
void *csr_asset_read_s( const char *path, i64 *size );
void *csr_asset_read( const char *path );
char *csr_textasset_read_s( const char *path, i64 *size );
char *csr_textasset_read( const char *name );

i64 fs_file_size( FILE *fileptr );

// Path handling
// -------------

// Find file path extension, returns NULL if no ext (0x00)
char *csr_findext( char *path, char const delim );

// gets rid of extension on string only left with (folder)+filename
void csr_stripext( char *path );

int csr_path_is_abs( char const *path );

// Convert windows paths to unix.. sortof ( \something\\blahblah .. ) -> /something/blahblah/
// Does not handle drive letters, idea is to increase windows compatibility will the other functions above
void csr_path_winunix( char *path );

// Implementation
//=======================================================================================================================

#ifdef CSR_EXECUTABLE

i64 fs_file_size( FILE *fileptr )
{
	fseek( fileptr, 0, SEEK_END );
	i64 fsize = ftell( fileptr );
	fseek( fileptr, 0, SEEK_SET );
	
	return fsize;
}

void *fs_disk_open_read( const char *path, int const reserve_end, i64 *size )
{
	FILE *f = fopen( path, "rb" );
	if( f )
	{
		i64 fsize = fs_file_size( f );
		void *buf = csr_malloc( fsize + reserve_end );
		
		if( buf )
		{
			// Invalid / corrupt read
			if( fread( buf, 1, fsize, f ) != fsize )
			{
				free( buf );
				buf = NULL;
			}
		}
		
		*size = fsize;
		
		fclose( f );
		return buf;
	}
	else
	{
		return NULL;
	}
}

char *fs_disk_load_text( const char *path, i64 *size )
{
	char *buf;
	i64 fsize;
	
	if( (buf = fs_disk_open_read( path, 1, &fsize )) )
	{
		buf[ fsize ] = 0x00;
		*size = fsize +1;
		
		return buf;
	}

	return NULL;
}

void *csr_asset_read_s( const char *path, i64 *size )
{
	return fs_disk_open_read( path, 0, size );
}

void *csr_asset_read( const char *path )
{
	i64 size;
	return fs_disk_open_read( path, 0, &size );
}

char *csr_textasset_read_s( const char *path, i64 *size )
{
	return fs_disk_load_text( path, size );
}

char *csr_textasset_read( const char *name )
{
	i64 size;
	return fs_disk_load_text( name, &size );
}

char *csr_findext( char *path, char const delim )
{
	char *c, *ptr;

	c = path;
	ptr = NULL;
	
	while( *c )
	{
		if( *c == delim )
		{
			ptr = c + 1;
		}
	
		c ++;
	}

	return ptr;
}

void csr_stripext( char *path )
{
	char *point, *start;
	
	// Skip folders
	if( !(start = csr_findext( path, '/' )) )
	{
		start = path;
	}
	
	if( (point = csr_findext( start, '.' )) )
	{
		if( point > path )
		{
			*(point-1) = 0x00;
		}
	}
}

void csr_path_winunix( char *path )
{
	char *idx, *wr;
	wr = idx = path;
	
	while( *idx )
	{
		if( *idx == '\\' )
		{
			*idx = '/';
		}
		
		if( idx > path )
		{
			if( *(idx -1) == '/' && *idx == '/') idx ++;
		}
		
		*( wr ++ ) = *idx;
		
		idx ++;
	}
	
	*wr = 0x00;
}

int csr_path_is_abs( char const *path )
{
#ifdef _WIN32
	if( strlen( path ) < 2 ) return 0;
	return path[1] == ':';
#else
	if( strlen( path ) < 1 ) return 0;
	return path[0] == '/';
#endif
}

#endif
