// This software is not affiliated with Valve Corporation
//   We are not affiliated, associated, authorized, endorsed by, or in any way officially 
//   connected with Valve Corporation, or any of its subsidiaries or its affiliates. 
// 
//   All trademarks are property of their respective owners

// Abstract Valve file system 
//=======================================================================================================================

// Initialize game directories / pakfile
void fs_set_gameinfo( const char *path );
void fs_exit(void);

// Read asset as binary in full. Will search VPK, then searchpaths, or return NULL if not found
char *valve_fs_get( const char *path );

// Implementation
//=======================================================================================================================

#ifdef VALVE_IMPLEMENTATION

struct valve_filesystem
{
	char gamedir[ 512 ];
	char exedir[ 512 ];
	char bindir[ 512 ];		// TODO: This is currently not set?
	
	VPKHeader *vpk;
	char **searchpaths;

	FILE *current_archive;
	u16 	current_idx;
} 
fs_global;

void fs_set_gameinfo( const char *path )
{
	struct valve_filesystem *fs = &fs_global;
	
	vdf_node *info = vdf_open_file( path );
	if( !info ) 
		return;	
	
	// Set gamedir
	strcpy( fs->gamedir, path );
	csr_path_winunix( fs->gamedir );
	*csr_findext( fs->gamedir, '/' ) = 0x00;
	
	// Set exe dir
	strcpy( fs->exedir, fs->gamedir );
	strcat( fs->exedir, "../" );
	
	// Get all search paths from file
	vdf_node *search_paths = vdf_next(vdf_next(vdf_next( info, "GameInfo", NULL ), "FileSystem", NULL ), "SearchPaths", NULL );
	
	kv_foreach( search_paths, "Game", kv_game )
	{
		if( kv_game[0] == '|' ) continue; //TODO: deal with engine replacements?? maybe??
	
		char *buf;
		if( csr_path_is_abs( kv_game ) )
		{
			buf = csr_malloc( strlen( kv_game ) + 2 );
			strcpy( buf, kv_game );
			strcat( buf, "/" );
		}
		else
		{
			buf = csr_malloc( strlen( fs->exedir ) + strlen( kv_game ) + 2 );
			strcpy( buf, fs->exedir );
			strcat( buf, kv_game );
			strcat( buf, "/" );
		}
		
		fs->searchpaths = csr_sb_reserve( fs->searchpaths, 1, sizeof( char * ) );
		fs->searchpaths[ csr_sb_count( fs->searchpaths ) ] = buf;
		csr_sb_use( fs->searchpaths );
	}
	
	vdf_free_r( info );
	
	// Look for pak01_dir
	char pack_path[512];
	for( int i = 0; i < csr_sb_count( fs->searchpaths ); i ++ )
	{
		strcpy( pack_path, fs->searchpaths[i] );
		strcat( pack_path, "pak01_dir.vpk" );
	
		if( (fs->vpk = (VPKHeader *)csr_asset_read( pack_path )) )
		{
			break;
		}
	}
	
	if( !fs->vpk )
	{
		log_error( "Could not locate pak01_dir.vpk in %i searchpaths. Stock models will not load!", csr_sb_count( fs->searchpaths ) );
	}

	log_info( "fs_info:\n" );
	log_info( "  gamedir: %s\n", fs->gamedir );
	log_info( "   exedir: %s\n", fs->exedir );
	log_info( "      bin: %s\n", fs->bindir );
	log_info( "     pack: %s\n", pack_path );
	log_info( "  searchpaths:\n" );
  
	for( int i = 0; i < csr_sb_count( fs->searchpaths ); i ++ )
	{
		log_info( "    '%s'\n", fs->searchpaths[i] );  
	}
}

void fs_exit(void)
{
	struct valve_filesystem *fs = &fs_global;

	for( int i = 0; i < csr_sb_count( fs->searchpaths ); i ++ )
	{
		free( fs->searchpaths[ i ] );
	}
	csr_sb_free( fs->searchpaths );
	fs->searchpaths = NULL;

	if( fs->vpk )
	{
		vpk_free( fs->vpk );
		fs->vpk = NULL;
	}
	
	if( fs->current_archive )
	{
		fclose( fs->current_archive );
		fs->current_archive = NULL;
	}
}

char *valve_fs_get( const char *path )
{
	struct valve_filesystem *fs = &fs_global;
	
	VPKDirectoryEntry *entry;
	char pak[ 533 ];

	if( fs->vpk )
	{
		if( (entry = vpk_find( fs->vpk, path )) )
		{
			if( entry->ArchiveIndex != fs->current_idx )
			{
				if( fs->current_archive )
				{
					fclose( fs->current_archive );
					fs->current_archive = NULL;
				}
				
				fs->current_idx = entry->ArchiveIndex;
			}
			
			if( !fs->current_archive )
			{
				sprintf( pak, "%scsgo/pak01_%03hu.vpk", fs->exedir, fs->current_idx );
				fs->current_archive = fopen( pak, "rb" );
				
				if( !fs->current_archive )
				{
					log_error( "Could not locate %s\n", pak );
					return NULL;
				}
			}
			
			char *filebuf = csr_malloc( entry->EntryLength );
			
			fseek( fs->current_archive, entry->EntryOffset, SEEK_SET );
			if( fread( filebuf, 1, entry->EntryLength, fs->current_archive ) == entry->EntryLength )
			{
				return filebuf;
			}
			else
			{
				free( filebuf );
				return NULL;
			}
		}
	}
	
	// Use physical searchpaths
	char path_buf[ 512 ];
	
	for( int i = 0; i < csr_sb_count( fs->searchpaths ); i ++ )
	{
		strcpy( path_buf, fs->searchpaths[ i ] );
		strcat( path_buf, path );
		
		char *filebuf;
		if( (filebuf = csr_asset_read( path_buf )) )
		{
			return filebuf;
		}
	}
	
	return NULL;
}

#endif
